import Expire, Storage

import Globals
from Persistence import Persistent
import Acquisition
from DateTime import DateTime
from OFS import content_types

from string import split
from webdav.common import rfc1123_date

class Cache(Persistent, Acquisition.Implicit):
    """ Cache superclass with support for different storage methods and 
        expiration algorithms.  It's mainly meant for caching data
        that is returned to the client - e.g. DTML Documents, XML Documents,
        etc.. 
    """
        
    __ac_permissions__=(('Update Cache', ('update',)),)

    def __init__(self):
        self.expireService = self._getExpireService()
        self.storageService = self._getStorageService()
        self._v_cache = None
        self.lastreadtime = 0
    
    def _getExpireService(self):
        # override this function if you want to substitute your own
        # expiration service, by returning an appropriate object
        
        return Expire.timed(delay=3600)
    
    def _getStorageService(self):
        # override this function if you want to substitute your own 
        # storage service, by returning an appropriate object
        
        return Storage.zodb()
        
    def __setstate__(self, state):
        """ initialize cache when object gets unpickled from zodb """
        Persistent.__setstate__(self, state)
        
        if not self.expireService.hasExpired():
            self._v_cache = self.storageService.load()
        else:
            # __str__ or __call__ will update cache when accessed,
            # so there's no need to do so ourselves
            # but we do need to create volatile vars
            self._v_cache = None
    
    def __str__(self, *args, **kwargs):
        """ return contents of cache when called by <dtml-var> """
        if hasattr(self, 'REQUEST'):
            REQUEST=self.REQUEST
        else:
            REQUEST={}

        if (self._v_cache is not None) and self.expireService.hasExpired():
            self._v_cache = None
        if self._v_cache is None:
            self._update(None, REQUEST, args, kwargs)
        return self._v_cache
    
    #Make sure we don't acquire index_html
    index_html=None
    
    def __call__(self, client=None, REQUEST={}, RESPONSE=None, *args, **kwargs):
        """ Return contents of cache when called directly by browser """
        
        ExpireService = self.expireService
        lastreadtime = self.lastreadtime
        
        # check if cache has expired and update if necessary                                                                          
        if (self._v_cache is not None) and ExpireService.hasExpired():
            self._v_cache = None
        if self._v_cache is None:
            self._update(client, REQUEST, args, kwargs)
        
        # should we send back a 304 Not Modified status?        
        ms=REQUEST.get_header('If-Modified-Since', None)
        if ms is not None:
            ms=split(ms, ';')[0]
            ms=DateTime(ms).timeTime()
            
            # CHECK IN NEXT VERSION:
            # Is the OFS.Image still broken?  And does this work on Windows?
            if lastreadtime <= ms:
                RESPONSE.setStatus(304)
                return RESPONSE
        
        # if we havn't sent a 304, send the content with appropiate headers
        RESPONSE.setHeader('Last-Modified', rfc1123_date(lastreadtime))
        sendExpire, expireTime = ExpireService.expireHeader()
        if sendExpire:
            RESPONSE.setHeader('Expires', rfc1123_date(expireTime))
        RESPONSE.setHeader('Content-Length', self.size)
        RESPONSE.write(self._v_cache)
        return ''
            
    def _update(self, client=None, REQUEST={}, args=(), kwargs={}):
        """ Force update of contents of cache """
        newCacheData = self._getData(client, REQUEST, args, kwargs)
        
        if newCacheData != self._v_cache:
            self.lastreadtime = int(DateTime().timeTime())
            self.expireService.doExpire(self.lastreadtime)
            self._v_cache = newCacheData
            self.size = len(self._v_cache)
            self.storageService.dump(self._v_cache)
            
    def update(self, REQUEST, **kwargs):
        """ Web interface to _update """
        self._update(None, REQUEST, (), kwargs)
    
    def _getData(self, client, REQUEST, args, kwargs):
        """ the all important function - return data to be stored in cache.
            should be implemented by subclasses. """
        raise NotImplemntedError

Globals.default__class_init__(Cache)
